from datetime import datetime
from enum import Enum, auto
from typing import Any, Dict, Optional

from pydantic import BaseModel


class JobStatus:
    inactive = auto()
    active = auto()
    failed = auto()
    finished = auto()


class Job(BaseModel):
    id: int
    payload: Dict[str, Any]
    try_count: int
    timeout: int
    max_tries: int
    state: str
    created_at: datetime
    updated_at: datetime
    priority: int
    result: Optional[Dict[str, Any]] = None



from datetime import datetime
from typing import Any, Dict, List, Union
import time
from psycopg import Connection, sql
from psycopg.rows import dict_row
from psycopg.types.json import Jsonb

from pqq import types


class Queue:
    def __init__(self, name: str, conn: Connection):
        self.name = name
        self.conn = conn
        self.conn.row_factory = dict_row
        self.conn.autocommit = False

    def create(self):
        txt = sql.SQL(
            f"CREATE TABLE IF NOT EXISTs {self.name}() INHERITS (base_queue);"
        )
        self.conn.execute(txt)
        self.conn.commit()

    def put(self, payload: Dict[str, Any]):
        txt = sql.SQL(
            "insert into {table} (payload) values (%s);".format(table=self.name)
        )
        self.conn.execute(txt, [Jsonb(payload)])
        self.conn.commit()

    def change_state(self, jobid: int, state: str):
        txt = sql.SQL(
            "UPDATE {table} SET state = %s, updated_at = %s where id = %s;".format(
                table=self.name
            )
        )
        now = datetime.utcnow()
        self.conn.execute(txt, [state, now, jobid])
        self.conn.commit()

    def get(self, block=True, timeout=10.0) -> types.Job:
        started = time.time()
        if block and timeout:
            waiting = True
            while waiting:
                rsp = self.get_nowait()
                if rsp:
                    return rsp
                elapsed = time.time() - started
                if elapsed > timeout:
                    raise TimeoutError()
                time.sleep(0.1)
        rsp = self.get_nowait()
        if not rsp:
            raise KeyError()
        return rsp

    def get_nowait(self) -> Union[types.Job, None]:
        now = datetime.utcnow()
        txt = sql.SQL(
            "UPDATE {table} SET state = 'active', updated_at = %s WHERE id = "
            "(select id from {table} where state = 'inactive'"
            " limit 1 for update skip locked) RETURNING *;".format(table=self.name)
        )
        j = None
        row = self.conn.execute(txt, [now]).fetchone()
        if row:
            j = types.Job(**row)
            # self.change_state(j.id, "active")
            # j.state = "active"
            self.conn.commit()
        return j

    def get_all(self):
        txt = sql.SQL("select * from {table};".format(table=self.name))
        rows = self.conn.execute(txt).fetchall()
        return rows

    def clean_failed(self):
        txt = sql.SQL("delete from {table} where state = %s;".format(table=self.name))
        self.conn.execute(txt, ["failed"])
        self.conn.commit()

    def clean_finished(self):
        txt = sql.SQL("delete from {table} where state = %s;".format(table=self.name))
        curr = self.conn.cursor()
        with self.conn.transaction():
            curr.execute(txt, ["finished"])
        self.conn.commit()

    def delete_queue(self):
        txt = sql.SQL("DROP TABLE {table};".format(table=self.name))
        curr = self.conn.cursor()
        with self.conn.transaction():
            curr.execute(txt)
        self.conn.commit()

    def clean(self):
        txt = sql.SQL("delete from {table};".format(table=self.name))
        curr = self.conn.cursor()
        with self.conn.transaction():
            curr.execute(txt)
        self.conn.commit()

    def delete_job(self, jobid: int):
        txt = sql.SQL("delete from {table} where id = %s;".format(table=self.name))
        curr = self.conn.cursor()
        with self.conn.transaction():
            curr.execute(txt, [jobid])
        self.conn.commit()

    def get_job(self, jobid: int) -> types.Job:
        txt = sql.SQL(
            "select * from {table} where id = %s limit 1;".format(table=self.name)
        )
        row = self.conn.execute(txt, [jobid]).fetchone()
        return types.Job(**row)

    def close(self):
        self.conn.close()



from enum import Enum
import psycopg
from psycopg.types.enum import EnumInfo, register_enum

# import signal


SELECT_TABLES = "SELECT table_schema,table_name FROM information_schema.tables WHERE table_schema = '{}' ORDER BY table_schema,table_name".format(
    "public"
)


def create_conn(dsn: str, autocommit=True) -> psycopg.Connection:
    conn = psycopg.connect(dsn, autocommit=autocommit)
    return conn


def async_create_conn(dsn, autocommit=True):
    aconn = psycopg.AsyncConnection.connect(dsn, autocommit=autocommit)
    return aconn


def run_migration(dsn, fp: str):
    conn = create_conn(dsn, autocommit=False)
    try:
        with open(fp, "r") as f:
            _ddl = f.read()
        conn.execute(_ddl)
    except BaseException:
        conn.rollback()
    else:
        conn.commit()
    finally:
        conn.close()


def register_db_enum(conn, sql_name: str, enum_type: Enum):
    enum = EnumInfo(conn, sql_name)
    register_enum(enum, conn, enum_type)


def select_tables(conn, schema="public"):
    stmt = SELECT_TABLES.format(schema)
    rows = conn.execute(stmt).fetchall()
    return rows
